unit FileObjUnit;

interface

{$I SxInclude.inc}

uses
  ComObj, ActiveX, ExMegaDemoServer_TLB, StdVcl, Graphics, AxCtrls, SxTools,
  SxIconList, SysUtils, Windows, Classes,
  {$IFNDEF LESS_THAN_D6}
  Variants,
  {$ENDIF}
  MainServerUnit;

type
  IFileObjInit=interface
  ['{14572229-480F-4DCA-B54B-09252F5FCABB}']
    procedure SetStorage(stg:IStorage; parent:IMDFileObj);
  end;

  TMDFileObj = class(TAutoObject, IMDFileObj, IFileObjInit)
  private
    FStorage: IStorage;
    FSubStorage:IStorage;
    FName:WideString;
    FReadOnly:Boolean;
    FIsSystemIcon:Boolean;
    FIcon:TIcon;
    FOpenIcon:TIcon;
    FKind:Integer;
    FParent: IMDFileObj;
  protected
    // IMDFileObj interface methods
    function Get_Name: WideString; safecall;
    procedure Set_Name(const Value: WideString); safecall;
    function Get_ReadOnly: WordBool; safecall;
    procedure Set_ReadOnly(Value: WordBool); safecall;
    function Get_IsSystemIcon: WordBool; safecall;
    procedure Set_IsSystemIcon(Value: WordBool); safecall;
    procedure GetIcon(out Icon: OleVariant); safecall;
    procedure GetOpenIcon(out OpenIcon: OleVariant); safecall;
    procedure SetIcon(Icon: OleVariant); safecall;
    procedure SetOpenIcon(OpenIcon: OleVariant); safecall;
    function Get_Kind: TFileObjKind; safecall;
    procedure Set_Kind(Value: TFileObjKind); safecall;
    // IFileObjInit interface methods
    procedure SetStorage(stg:IStorage; parent:IMDFileObj);
    procedure UpdateChanges; safecall;
    function Get_ChildCount: Integer; safecall;
    procedure GetChild(Index: Integer; out pChild: IMDFileObj); safecall;
    procedure GetChildByName(const pName: WideString; out pChild: IMDFileObj);
      safecall;
    procedure GetChildNames(out pNames: WideString); safecall;
    procedure GetChildByRelativePath(const pPath: WideString;
      out pChild: IMDFileObj); safecall;
    procedure AddNewChild(const pName: WideString; out pChild: IMDFileObj);
      safecall;
    function RenameChild(const pOldName, pNewName: WideString): HResult;
      safecall;
    procedure DeleteChild(const pChildName: WideString); safecall;
    procedure SetContents(pData: OleVariant); safecall;
    function GetContents(out pData: OleVariant): HResult; safecall;
    function Get_Size: Integer; safecall;
    { Protected declarations }
  public
    procedure Initialize;override;
    destructor Destroy;override;
    procedure ReadSettings;
    procedure GetSubStorage(out SubStg:IStorage);
    property Storage:IStorage read FStorage;
    property Parent:IMDFileObj read FParent;
  end;

const
  nameSettings='Settings';
  nameFileStorage='FileStorage';
  nameSubStorage='SubStorage';
  nameDataStream='DataStream';
  stgMode=STGM_READWRITE or STGM_SHARE_EXCLUSIVE or STGM_DIRECT;

implementation

uses ComServ;

function TMDFileObj.Get_Name: WideString;
begin
  Result:=FName;
end;

procedure TMDFileObj.Set_Name(const Value: WideString);
begin
  FName:=Value;
end;

function TMDFileObj.Get_ReadOnly: WordBool;
begin
  Result:=FReadOnly;
end;

procedure TMDFileObj.Set_ReadOnly(Value: WordBool);
begin
  FReadOnly:=Value;
end;

function TMDFileObj.Get_IsSystemIcon: WordBool;
begin
  Result:=FIsSystemIcon;
end;

procedure TMDFileObj.Set_IsSystemIcon(Value: WordBool);
begin
  FIsSystemIcon:=Value;
end;

procedure TMDFileObj.GetIcon(out Icon: OleVariant);
var
  tempStream:TMemoryStream;
begin
  tempStream:=TMemoryStream.Create;
  try
    FIcon.SaveToStream(tempStream);
    Icon:=SxStreamToVarArray(tempStream);
  finally
    FreeAndNil(tempStream);
  end;
end;

procedure TMDFileObj.GetOpenIcon(out OpenIcon: OleVariant);
var
  tempStream:TMemoryStream;
begin
  tempStream:=TMemoryStream.Create;
  try
    FIcon.SaveToStream(tempStream);
    OpenIcon:=SxStreamToVarArray(tempStream);
  finally
    FreeAndNil(tempStream);
  end;
end;

procedure TMDFileObj.SetIcon(Icon: OleVariant);
var
  tempStream:TMemoryStream;
begin
  tempStream:=TMemoryStream.Create;
  try
    SxVarArrayToStream(Icon,tempStream);
    tempStream.Position:=0;
    FIcon.LoadFromStream(tempStream);
  finally
    FreeAndNil(tempStream);
  end;
end;

procedure TMDFileObj.SetOpenIcon(OpenIcon: OleVariant);
var
  tempStream:TMemoryStream;
begin
  tempStream:=TMemoryStream.Create;
  try
    SxVarArrayToStream(OpenIcon,tempStream);
    tempStream.Position:=0;
    FOpenIcon.LoadFromStream(tempStream);
  finally
    FreeAndNil(tempStream);
  end;
end;

function TMDFileObj.Get_Kind: TFileObjKind;
begin
  Result:=FKind;
end;

procedure TMDFileObj.Set_Kind(Value: TFileObjKind);
begin
  FKind:=Value;
end;

procedure TMDFileObj.SetStorage(stg: IStorage; parent:IMDFileObj);
begin
  FStorage:=stg;
  FParent:=parent;
  ReadSettings;
end;

procedure TMDFileObj.Initialize;
begin
  inherited;
  FStorage:=nil;
  FIcon:=TIcon.Create;
  FOpenIcon:=TIcon.Create;
end;

procedure TMDFileObj.UpdateChanges;
var
  tempSettings:IStream;
  tempStream:TOleStream;
  SxStream:TSxStream;
  tempIconList:TSxIconList;

  tempIcon:TIcon;

begin
  if Storage=nil then
     Exit;
  if Failed(Storage.CreateStream(nameSettings,stgMode,0,0,tempSettings)) then
     if Failed(Storage.OpenStream(nameSettings,nil,stgMode,0,tempSettings)) then
        raise Exception.Create('Failed to create attributes storage for element')
     else
        tempSettings.SetSize(0);
  tempStream:=TOleStream.Create(tempSettings);
  try
    SxStream:=TSxStream.Create(tempStream);
    try
      SxStream.WriteWideString(FName);
      SxStream.WriteBoolean(FReadOnly);
      SxStream.WriteBoolean(FIsSystemIcon);
      SxStream.WriteInteger(FKind);
      tempIconList:=TSxIconList.Create(nil);
      try
        if not FIcon.Empty then
        begin
          tempIconList.AddIcon(FIcon.Handle);
          tempIconList.AddIcon(FOpenIcon.Handle);
        end;
        tempIconList.SaveToStream(tempStream);
      finally
        FreeAndNil(tempIconList);
      end;
    finally
      FreeAndNil(SxStream);
    end;
  finally
    FreeAndNil(tempStream);
  end;
  tempSettings:=nil;
  ReadSettings;
end;

procedure TMDFileObj.ReadSettings;
var
  tempSettings:IStream;
  tempStream:TOleStream;
  SxStream:TSxStream;
  tempIconList:TSxIconList;
  tempIcon:TIcon;
begin
  if Storage=nil then
     Exit;
  if Failed(Storage.OpenStream(nameSettings,nil,stgMode,0,tempSettings)) then
  begin
    if Failed(Storage.CreateStream(nameSettings,stgMode,0,0,tempSettings)) then
       raise Exception.Create('Failed to create attributes storage for element');
    Exit;
  end;
  tempStream:=TOleStream.Create(tempSettings);
  try
    SxStream:=TSxStream.Create(tempStream);
    try
      FName:=SxStream.ReadWideString;
      FReadOnly:=SxStream.ReadBoolean;
      FIsSystemIcon:=SxStream.ReadBoolean;
      FKind:=SxStream.ReadInteger;
      tempIconList:=TSxIconList.Create(nil);
      try
        tempIconList.LoadFromStream(tempStream);
        if tempIconList.IconCount<2 then
           FIsSystemIcon:=True
        else
        begin
          FIcon.Handle:=tempIconList.GetHIcon(0);
          FOpenIcon.Handle:=tempIconList.GetHIcon(1);

          tempIcon:=TIcon.Create;
          tempIcon.Width:=32;
          tempIcon.Height:=32;
          tempIcon.Handle:=tempIconList.GetHIcon(0);
          MDServerWindow.Image1.Picture.Icon.Assign(tempIcon);
          FreeAndNil(tempIcon);

          MDServerWindow.Image2.Picture.Icon.Assign(FIcon);
        end;
      finally
        FreeAndNil(tempIconList);
      end;
    finally
      FreeAndNil(SxStream);
    end;
  finally
    FreeAndNil(tempStream);
  end;
end;

destructor TMDFileObj.Destroy;
begin
  FStorage:=nil;
  FreeAndNil(FIcon);
  FreeAndNil(FOpenIcon);
  inherited;
end;

function TMDFileObj.Get_ChildCount: Integer;
var
  tempStorage:IStorage;
  tempEnum:IEnumStatStg;
  StatStg:TStatStg;
  Fetched:Longint;
begin
  Result:=0;
  GetSubStorage(tempStorage);
  if tempStorage=nil then
     Exit;
  if Failed(tempStorage.EnumElements(0,nil,0,tempEnum)) then
     Exit;
  if tempEnum=nil then
     Exit;
  Result:=0;
  while tempEnum.Next(1,StatStg,@Fetched)=S_OK do
  begin
    if Fetched<1 then
       Break;
    Result:=Result+1;
  end;
end;

procedure TMDFileObj.GetChild(Index: Integer; out pChild: IMDFileObj);
var
  tempStorage:IStorage;
  tempEnum:IEnumStatStg;
  statStg:TStatStg;
  fetched:Longint;
  childInit:IFileObjInit;
  tempLocatedStg:IStorage;
begin
  pChild:=nil;
  GetSubStorage(tempStorage);
  if tempStorage=nil then
     Exit;
  if Failed(tempStorage.EnumElements(0,nil,0,tempEnum)) then
     Exit;
  if tempEnum=nil then
     Exit;
  while tempEnum.Next(1,StatStg,@Fetched)=S_OK do
  begin
    if Fetched<1 then
       Break;
    if Index=0 then
    begin
      if Failed(tempStorage.OpenStorage(StatStg.pwcsName,nil,stgMode,nil,0,tempLocatedStg)) then
         Break;
      pChild:=CoMDFileObj.Create;
      childInit:=pChild as IFileObjInit;
      childInit.SetStorage(tempLocatedStg, Self);
    end
    else
      Index:=Index-1;
  end;
end;

procedure TMDFileObj.GetSubStorage(out SubStg: IStorage);
var
  tempStorage:IStorage;
  res:HRESULT;
begin
  SubStg:=nil;
  if FSubStorage=nil then
  begin
    Res:=Storage.OpenStorage(nameSubStorage,nil,stgMode,nil,0,tempStorage);
    if Failed(Res) then
    begin
      Res:=Storage.CreateStorage(nameSubStorage,stgMode,0,0,tempStorage);
      if Failed(Res) then
         OleCheck(Res);
    end;
    FSubStorage:=tempStorage;
  end;
  SubStg:=FSubStorage;
end;

procedure TMDFileObj.GetChildByName(const pName: WideString;
  out pChild: IMDFileObj);
var
  tempStorage:IStorage;
  tempTargetStg:IStorage;
  childInit:IFileObjInit;
begin
  pChild:=nil;
  GetSubStorage(tempStorage);
  if tempStorage=nil then
     Exit;
  if Failed(tempStorage.OpenStorage(PWideChar(pName),nil,stgMode,nil,0,tempTargetStg)) then
     Exit;
  pChild:=CoMDFileObj.Create;
  childInit:=pChild as IFileObjInit;
  childInit.SetStorage(tempTargetStg, Self);
end;

procedure TMDFileObj.GetChildNames(out pNames: WideString);
var
  tempStorage:IStorage;
  tempEnum:IEnumStatStg;
  StatStg:TStatStg;
  Fetched:Longint;
  tempStr:TStrings;
begin
  tempStr:=TStringList.Create;
  try
    GetSubStorage(tempStorage);
    if tempStorage=nil then
       Exit;
    if Failed(tempStorage.EnumElements(0,nil,0,tempEnum)) then
       Exit;
    if tempEnum=nil then
       Exit;
    while tempEnum.Next(1,StatStg,@Fetched)=S_OK do
    begin
      if Fetched<1 then
         Break;
      tempStr.Add(WideCharToString(StatStg.pwcsName));
    end;
    pNames:=tempStr.Text;
  finally
    FreeAndNil(tempStr);
  end;
end;

procedure TMDFileObj.GetChildByRelativePath(const pPath: WideString;
  out pChild: IMDFileObj);
var
  tempStr:TStrings;
  i:Integer;
  tempParent:IMDFileObj;
  tempChild:IMDFileObj;
begin
  pChild:=nil;
  tempStr:=TStringList.Create;
  try
    FilePathToStrings(pPath,tempStr);
    tempParent:=Self;
    for i:=0 to tempStr.Count-1 do
    begin
      tempParent.GetChildByName(tempStr[i],tempChild);
      if tempChild=nil then
         Break;
      tempParent:=tempChild;
    end;
    pChild:=tempChild;
  finally
    FreeAndNil(tempStr);
  end;
end;

procedure TMDFileObj.AddNewChild(const pName: WideString;
  out pChild: IMDFileObj);
var
  tempStorage:IStorage;
  tempTargetStg:IStorage;
  childInit:IFileObjInit;
begin
  pChild:=nil;
  GetSubStorage(tempStorage);
  if tempStorage=nil then
     Exit;
  if Succeeded(tempStorage.OpenStorage(PWideChar(pName),nil,stgMode,nil,0,tempTargetStg)) then
     Exit
  else
  if Failed(tempStorage.CreateStorage(PWideChar(pName),stgMode,0,0,tempTargetStg)) then
     Exit;
  pChild:=CoMDFileObj.Create;
  childInit:=pChild as IFileObjInit;
  childInit.SetStorage(tempTargetStg, Self);
  pChild.Name:=pName;
  pChild.Kind:=okFolder;
  pChild.UpdateChanges;
end;

function TMDFileObj.RenameChild(const pOldName,
  pNewName: WideString): HResult;
var
  tempStorage:IStorage;
  tempTargetStg:IStorage;
  tempChild:IMDFileObj;
  tempChildInit:IFileObjInit;
begin
  Result:=E_FAIL;
  GetSubStorage(tempStorage);
  if tempStorage=nil then
     Exit;
  Result:=tempStorage.RenameElement(PWideChar(pOldName),PWideChar(pNewName));
  if Succeeded(Result) then
  begin
    Result:=tempStorage.OpenStorage(PWideChar(pNewName),nil,stgMode,nil,0,tempTargetStg);
    if Failed(Result) then
       Exit;
    tempChild:=CoMDFileObj.Create;
    tempChildInit:=tempChild as IFileObjInit;
    tempChildInit.SetStorage(tempTargetStg,Self);
    tempChild.Name:=pNewName;
    tempChild.UpdateChanges;
  end;
end;

procedure TMDFileObj.DeleteChild(const pChildName: WideString);
var
  tempStorage:IStorage;
begin
  GetSubStorage(tempStorage);
  if tempStorage=nil then
     Exit;
  tempStorage.DestroyElement(PWideChar(pChildName));
end;

procedure TMDFileObj.SetContents(pData: OleVariant);
var
  dataStream:IStream;
  tempOleStream:TOleStream;
begin
  if Failed(Storage.OpenStream(nameDataStream,nil,stgMode,0,dataStream)) then
     if Failed(Storage.CreateStream(nameDataStream,stgMode,0,0,dataStream)) then
        Exit;
  dataStream.SetSize(0);
  tempOleStream:=TOleStream.Create(dataStream);
  try
    SxVarArrayToStream(pData,tempOleStream);
  finally
    FreeAndNil(tempOleStream);
  end;
end;

function TMDFileObj.GetContents(out pData: OleVariant): HResult;
var
  dataStream:IStream;
  tempOleStream:TOleStream;
begin
  pData:=null;
  Result:=Storage.OpenStream(nameDataStream,nil,stgMode,0,dataStream);
  if Failed(Result) then
     Exit;
  tempOleStream:=TOleStream.Create(dataStream);
  try
    pData:=SxStreamToVarArray(tempOleStream);
  finally
    FreeAndNil(tempOleStream);
  end;
end;

function TMDFileObj.Get_Size: Integer;
var
  dataStream:IStream;
  tempOleStream:TOleStream;
  res:HRESULT;
begin
  Result:=0;
  res:=Storage.OpenStream(nameDataStream,nil,stgMode,0,dataStream);
  if Failed(res) then
     Exit;
  tempOleStream:=TOleStream.Create(dataStream);
  try
    Result:=tempOleStream.Size;
  finally
    FreeAndNil(tempOleStream);
  end;
end;

initialization
  TAutoObjectFactory.Create(ComServer, TMDFileObj, Class_MDFileObj,
    ciMultiInstance, tmApartment);
end.
